/*
 *      cluster.c -- find and rank clusters by size, graph largest
 *
 *      Copyright (c) 2011 Dave Raggett
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>   // for strlen
#include <strings.h>
#include <sqlite3.h>

#include "hashtable.h"
#include "rbtree.h"
#include "analyst.h"
#include "top_sites.h"

// bit flags for nodes for marking visits etc
#define DASHBOT_VISITED 1
#define DASHBOT_GRAPHED 2

static struct hashtable *sites = NULL;
static struct _dashbot_node *all_nodes = NULL;
static RbTree *clusters = NULL;

typedef struct _dashbot_node {
  char *host;
  int flags;
  RbTree *in;
  RbTree *out;
  struct _dashbot_node *next;
} DashbotNode;

typedef struct _dashbot_node_list {
  DashbotNode *node;
  struct _dashbot_node_list *next;
} DashbotNodeList;

typedef struct _dashbot_arc {
  DashbotNode *node;
  int offsite;
} DashbotArc;

typedef struct _dashbot_cluster {
  DashbotNode *root;
  int size;  // number of nodes in cluster
  struct _dashbot_cluster *next;  // peers with same size
} DashbotCluster;


// **** list routines for debugging ****
static void list_link(RbKey key, RbValue value, RbContext context)
{
  //DashbotNode *node = (DashbotNode *)key;
  DashbotArc *arc = (DashbotArc *)value;
  printf("  -- %s (%d)\n", arc->node->host, arc->offsite);
}

static void list_node(DashbotNode *node)
{
  printf("%s: in %ld, out %ld\n", node->host,
          rbTreeSize(node->in), rbTreeSize(node->out));

  rbTreeApplyLR(node->in, (RbApplyFn)list_link, (RbContext)null);
  rbTreeApplyLR(node->out, (RbApplyFn)list_link, (RbContext)null);
}

static void list_nodes()
{
  DashbotNode *node = all_nodes;

  while (node)
  {
    list_node(node);
    node = node->next;
  }
}



static void new_arc(RbNode *node, RbKey key, RbValue value)
{
  node->key = key;
  node->value = (RbValue)value;
}

static void update_arc(RbNode *node, RbValue value)
{
  assert(0);
  if (node->value)
    free(node->value);

  node->value = (RbValue)value;
}

// for freeing data stored in red-black tree node
static void  erase_arc(RbKey key, RbValue value)
{
  if (value)
    free(value);
}

static int compare_arcs(RbKey key1, RbKey key2)
{
  DashbotNode *node1 = (DashbotNode *)key1;
  DashbotNode *node2 = (DashbotNode *)key2;
  return strcasecmp(node1->host, node2->host);
}

static void add_link(DashbotNode *host, DashbotNode *third_pty, int offsite)
{
  DashbotArc *arc;

  arc = (DashbotArc *)malloc(sizeof(DashbotArc));
  assert(arc);
  arc->node = third_pty;
  arc->offsite = offsite;

  rbTreeInsertKey(host->out, (RbKey)third_pty, (RbValue)arc);

  arc = (DashbotArc *)malloc(sizeof(DashbotArc));
  assert(arc);
  arc->node = host;
  arc->offsite = offsite;

  rbTreeInsertKey(third_pty->in, (RbKey)host, (RbValue)arc);
}

static DashbotNode *create_node(char *host)
{
  DashbotNode *node = (DashbotNode *)calloc(1, sizeof(DashbotNode));
  assert(node);

  assert(host);
  node->host = strdup(host);
  assert(node->host);
  node->in = rbTreeNew(new_arc, update_arc, erase_arc, compare_arcs);
  node->out = rbTreeNew(new_arc, update_arc, erase_arc, compare_arcs);
  node->next = all_nodes;
  all_nodes = node;
  return node;
}

static void free_node(DashbotNode *node)
{
  if (node)
  {
    free(node->host);
    rbTreeFree(node->in);
    rbTreeFree(node->out);
    free(node);
  }
}

static int free_all_nodes()
{
  DashbotNode *next, *node = all_nodes;
  int count = 0;

  while (node)
  {
    next = node->next;
    free_node(node);
    ++count;
    node = next;
  }

  return count;
}

static DashbotNode *get_node(char *host)
{
  DashbotNode *node = (DashbotNode *)hashtable_search(sites, host);

  if (node == null)
  {
    node = create_node(host);
    assert(hashtable_insert(sites, node->host, node, 2));
  }

  return node;
}

int check_missing_sites(sqlite3 *db, char *name)
{
  int i, missing = 0;
  SiteDesc *desc;
  DashbotNode *node;
  char *domain;
  char host[1024];
  FILE *fp = fopen(name, "w+");

  if (!fp)
  {
    fprintf(stderr, "Can't write to file %s\n", name);
    sqlite3_close(db);
    exit(1);
  }

  fprintf(fp, "\nvar missing_sites = [\n");

  for (i = 0; i < NUM_SITES; ++i)
  {
    desc = top_sites + i;
    domain = desc->domain;

    node = (DashbotNode *)hashtable_search(sites, domain);

    if (node)
      continue;

    strcpy(host, "www.");
    strcat(host, domain);

    node = (DashbotNode *)hashtable_search(sites, host);

    if (!node)
    {
      ++missing;
      fprintf(fp, "  \"%s\",\n", host);
    }
  }

  fprintf(fp, "];\n\n");
  return missing;
}

// insert host into hashable as node
static int note_sites(void *not_used, int argc, char **argv, char **azColName)
{
  char *host = argv[0];
  get_node(host);

  return 0;
}

// should extend node structure to collect
// other columns from the site_info table
static void examine_sites(sqlite3 *db)
{
  char sql[1024];
  char *zErrMsg = 0;

  sprintf(sql, "SELECT host FROM site_info");

  int rc = sqlite3_exec(db, sql, note_sites, null, &zErrMsg);

  if( rc != SQLITE_OK )
  {
    fprintf(stderr, "SQL error: %s\n", zErrMsg);
    sqlite3_free(zErrMsg);
  }
}

// insert host into hashable as node
static int note_links(void *not_used, int argc, char **argv, char **azColName)
{
  char *page_host = argv[0];
  char *third_party = argv[1];
  char *offsite = argv[2];
  int flag = 0;

  if (page_host && third_party)
  {
    DashbotNode *host_node = get_node(page_host);
    DashbotNode *third_pty_node = get_node(third_party);

    sscanf(offsite, "%d", &flag);
    add_link(host_node, third_pty_node, flag);
  }
  else
    fprintf(stderr, "Warning: missing domains -- host: %s, 3rd: %s, offset: %s\n", page_host, third_party, offsite);

  return 0;
}

// should extend node structure to collect
// other columns from the site_info table
static void examine_third_parties(sqlite3 *db, bool indirect)
{
  char *sql;
  char *zErrMsg = 0;

  if (indirect)
    sql = "SELECT page_host,third_party,offsite FROM parties";
  else
    sql = "SELECT parent,child,offsite FROM relations";

  int rc = sqlite3_exec(db, sql, note_links, null, &zErrMsg);

  if( rc != SQLITE_OK )
  {
    fprintf(stderr, "SQL error: %s\n", zErrMsg);
    sqlite3_free(zErrMsg);
  }
}

static void new_cluster(RbNode *node, RbKey key, RbValue value)
{
  node->key = (RbValue)value;
  node->value = (RbValue)value;
}

static void update_cluster(RbNode *node, RbValue value)
{
  DashbotCluster *cluster1 = (DashbotCluster *)node->value;
  DashbotCluster *cluster2 = (DashbotCluster *)value;

  cluster2->next = cluster1->next;
  cluster1->next = cluster2;
}

// for freeing data stored in red-black tree node
static void  erase_cluster(RbKey key, RbValue value)
{
  DashbotCluster *cluster, *next;

  if (value)
  {
    cluster = (DashbotCluster *)value;

    while (cluster)
    {
      next = cluster->next;
      free(cluster);
      cluster = next;
    }
  }
}

static int compare_clusters(RbKey key1, RbKey key2)
{
  DashbotCluster *cluster1 = (DashbotCluster *)key1;
  DashbotCluster *cluster2 = (DashbotCluster *)key2;
  int count1 = cluster1->size;
  int count2 = cluster2->size;

  return  count1 == count2 ? 0 : (count1 < count2 ? -1 : 1);
}

static DashbotCluster *create_cluster(DashbotNode *root)
{
  DashbotCluster *cluster = (DashbotCluster *)calloc(1, sizeof(DashbotCluster));
  assert(cluster);

  cluster->root = root;
  return cluster;
}

static bool visited_node(DashbotNode *node)
{
  return (node->flags & DASHBOT_VISITED) ? true : false;
}

static void mark_visited(DashbotNode *node)
{
  node->flags |= DASHBOT_VISITED;
}

// forward reference
static void trace_cluster(DashbotCluster *cluster, DashbotNode *node);

static void trace_arc(RbKey key, RbValue value, RbContext context)
{
  DashbotCluster *cluster = (DashbotCluster *)context;
  DashbotArc *arc = (DashbotArc *)value;
  trace_cluster(cluster, arc->node);
}

static void trace_cluster(DashbotCluster *cluster, DashbotNode *node)
{
  if (!visited_node(node))
  {
    cluster->size++;
    mark_visited(node);

    rbTreeApplyLR(node->in, (RbApplyFn)trace_arc, (RbContext)cluster);
    rbTreeApplyLR(node->out, (RbApplyFn)trace_arc, (RbContext)cluster);
  }
}

static void visit_cluster(RbKey key, RbValue value, RbContext context)
{
  DashbotCluster *cluster = (DashbotCluster *)value, *c;
  int count = 0;

  for (c = cluster; c; c = c->next)
  {
    ++count;
  }

  printf("  %d: %d (%d entries)\n", cluster->size, count * cluster->size, count);
}

static void examine_clusters()
{
  DashbotNode *node;
  DashbotCluster *cluster;
  long int max = 0, count = 0;

  // initialize global variable as red-black tree or sorted clusters
  clusters = rbTreeNew(new_cluster, update_cluster, erase_cluster, compare_clusters);

  // use top sites as nice nodes for cluster roots
  // this isn't necessary, but ...

  // now iterate through all nodes in case we missed some clusters
  for (node = all_nodes; node; node = node->next)
  {
    if (visited_node(node))
      continue;

    cluster = create_cluster(node);
    trace_cluster(cluster, node);

    ++count;

    if (max < cluster->size)
      max = cluster->size;

    printf("found cluster size %d root %s\n", cluster->size, cluster->root->host);

    rbTreeInsertKey(clusters, (RbKey)cluster, (RbValue)cluster);
  }

  printf("found %ld clusters in %ld sizes, max size is %ld\n", count, rbTreeSize(clusters), max);
  rbTreeApplyRL(clusters, (RbApplyFn)visit_cluster, (RbContext)null);
}

// local declarations for graphing a cluster
typedef struct {
  char *host_id;
  FILE *fp;
} GraphContext;

static void graph_node(FILE *fp, DashbotNode *node);

// write at most max bytes including null terminator
static bool format_node_id(DashbotNode *node, char *id, size_t max)
{
  int c;
  char *p = id, *q = node->host, *r = p + max - 1;

  if (p >= r)
    return false;

  *p++ = 's';
  *p++ = '_';


  while (p < r && (c = *q++))
  {
    *p++ = (c == '.'  || c == '-') ? '_' : c;
  }

  if (p < r)
  {
    *p = '\0';
    return true;
  }

  return false;
}

static void graph_arc1(RbKey key, RbValue value, RbContext context)
{
  GraphContext *gc = (GraphContext *)context;
  DashbotArc *arc = (DashbotArc *)value;
  char node_id[1024];

  format_node_id(arc->node, node_id, 1024);

  if (arc->offsite == 1)
    fprintf(gc->fp, "    %s -> %s [style=dotted];\n", gc->host_id, node_id);
  else
    fprintf(gc->fp, "    %s -> %s;\n", gc->host_id, node_id);
}

static void graph_arc2(RbKey key, RbValue value, RbContext context)
{
  DashbotArc *arc = (DashbotArc *)value;
  graph_node((FILE *)context, arc->node);
}

static void graph_node(FILE *fp, DashbotNode *node)
{

  if (!(node->flags & DASHBOT_GRAPHED))
  {
    char node_id[1024];
    node->flags |= DASHBOT_GRAPHED;
    format_node_id(node, node_id, 1024);

    fprintf(fp, "  %s [shape=\"ellipse\",label=\"%s\",color=black];\n",
            node_id, node->host);

    GraphContext gc;
    gc.host_id = node_id;
    gc.fp = fp;

    // this outputs the graphviz arcs
    rbTreeApplyLR(node->out, (RbApplyFn)graph_arc1, (RbContext)&gc);

    // this are used to ensure we trace all nodes in cluster
    rbTreeApplyLR(node->in, (RbApplyFn)graph_arc2, (RbContext)fp);
    rbTreeApplyLR(node->out, (RbApplyFn)graph_arc2, (RbContext)fp);
  }
}

static bool graph_cluster(DashbotCluster *cluster, char *name)
{
  FILE *fp = fopen(name, "w+");

  if (!fp)
  {
    fprintf(stderr, "Can't write to file %s\n", name);
    return false;
  }

  fprintf(fp, "digraph cluster {\n  graph [splines=true,overlap=false]\n");

  graph_node(fp, cluster->root);

  fputs("}\n", fp);
  fclose(fp);
  return true;
}

static void new_link(RbNode *node, RbKey key, RbValue value)
{
  node->key = key;
  node->value = (RbValue)value;
}

static void update_link(RbNode *node, RbValue value)
{
  DashbotNodeList *nl1 = (DashbotNodeList *)node->value;
  DashbotNodeList *nl2 = (DashbotNodeList *)value;

  nl2->next = nl1->next;
  nl1->next = nl2;
}

// for freeing data stored in red-black tree node
static void  erase_link(RbKey key, RbValue value)
{
  DashbotNodeList *nl, *next;

  if (value)
  {
    nl = (DashbotNodeList *)value;

    while (nl)
    {
      next = nl->next;
      free(nl);
      nl = next;
    }
  }
}

static int compare_links(RbKey key1, RbKey key2)
{
  long int count1 = (long int )key1;
  long int count2 = (long int)key2;
  return count1 == count2 ?  0 : (count1 < count2 ? 1 : -1);
}

typedef struct {
  sqlite3 *db;
  FILE *fp;
} LinkContext;

static void list_links(RbKey key, RbValue value, RbContext context)
{
  long int size = (long int)key;
  DashbotNodeList *item = (DashbotNodeList *)value;
  LinkContext *info = (LinkContext *)context;
  char sql[1024];
  char *zErrMsg = 0;
  sqlite3_stmt *statement;
  int count = 0;

  while (item)
  {
    ++count;
    item = item->next;
  }

  printf(" %ld:  (%d)\n", size, count);

  fprintf(info->fp,
          "      <tr>\n"
          "        <th colspan='18'>%ld sites with %d references</th>\n"
          "      </tr>\n", size, count);

  item = (DashbotNodeList *)value;

  // should move select statement outside of loop
  // with sqlite3_bind and sqlite3_reset inside loop

  while (item)
  {
    printf("   %s\n", item->node->host);

    snprintf(sql, 1024, "SELECT * FROM site_info WHERE host = \"%s\"", item->node->host);

    if (sqlite3_prepare_v2(info->db, sql, -1, &statement, NULL) == SQLITE_OK)
    {
      int cols = sqlite3_column_count(statement);
      assert(cols >= 19);

      while (sqlite3_step(statement) == SQLITE_ROW)
      {
        int i;
        char *value;

        fprintf(info->fp, "<tr>\n");

        for (i = 0; i <19; ++i)
        {
          if (i == 1 || i == 2 || i == 20)
            continue;

          value = (char*)sqlite3_column_text(statement, i);
          fprintf(info->fp, "<td>%s</td>\n", value);
        }

        fprintf(info->fp, "<tr>\n");
      }

      sqlite3_finalize(statement);
    }
    else
    {
      fprintf(stderr, "SQL error: %s\n", zErrMsg);
      sqlite3_free(zErrMsg);
      break;
    }

    item = item->next;
  }
}

// create a sorted list of nodes by diminishing
// number of references to the node as a 3rd party
static RbTree *sort_third_parties(sqlite3 *db)
{
  long int size;
  DashbotNode *node = all_nodes;
  RbTree *node_list = rbTreeNew(new_link, update_link, erase_link, compare_links);

  while (node)
  {
    size = (int)rbTreeSize(node->in);
    DashbotNodeList *item = (DashbotNodeList *)calloc(1, sizeof(DashbotNodeList));
    assert(item);
    item->node = node;

    rbTreeInsertKey(node_list, (RbKey)size, (RbValue)item);
    node = node->next;
  }

  printf("found %ld number of nodes grouped by 3rd party references to them\n",
         rbTreeSize(node_list));

  char *name = "sites.html";

  FILE *fp = fopen(name, "w+");

  if (fp)
  {
    LinkContext info;
    info.db = db;
    info.fp = fp;

    fprintf(fp,
            "<html>\n"
            "  <head>\n"
            "    <title>Third Party Sites</title>\n"
            "  </head>\n"
            "  <body>\n"
            "    <table border='1'>\n"
            "      <caption>Third Party Sites</caption>\n"
            "    <thead>\n"
            "      <tr>\n"
            "        <th>Site</th>\n"
            "        <th>Session Cookies</th>\n"
            "        <th>Lasting Cookies</th>\n"
            "        <th>Flash Cookies</th>\n"
            "        <th>Internal 3rd Parties</th>\n"
            "        <th>External 3rd Parties</th>\n"
            "        <th>Internal 3rd Party Session Cookies</th>\n"
            "        <th>Internal 3rd Party Lasting Cookies</th>\n"
            "        <th>Internal 3rd Party Flash Cookies</th>\n"
            "        <th>External rd Party Session Cookies</th>\n"
            "        <th>External 3rd Party Lasting Cookies</th>\n"
            "        <th>External 3rd Party Flash Cookies</th>\n"
            "        <th>DOM Storage</th>\n"
            "        <th>HTML5 pings</th>\n"
            "        <th>Invisible Images</th>\n"
            "        <th>Suspicious URLs</th>\n"
            "        <th>Geolocation Permission</th>\n"
            "        <th>P3P policy</th>\n"
            "      </tr>\n"
            "    </thead>\n"
            "    <tbody>\n");

    rbTreeApplyLR(node_list, (RbApplyFn)list_links, (RbContext)&info);

    fprintf(fp,
            "    </tbody>\n"
            "  </body>\n"
            "</html>\n");

    fclose(fp);
  }
  else
    fprintf(stderr, "couldn't open %s for writing\n", name);

  return node_list;
}

static int compare_strings(void *key1, void *key2)
{
  return !strcmp((char *)key1, (char *)key2);
}

size_t study_clusters(sqlite3 *db)
{
  sites = create_hashtable(8192, oat_hash2, compare_strings);

  if (!sites)
  {
    fprintf(stderr, "Can't create hash table\n");
    sqlite3_close(db);
    exit(1);
  }

  // first use site_info table to populate node dictionary
  // to catch websites that don't have any third parties
  examine_sites(db);
  check_missing_sites(db, "missing.js");

  // next use parties table to collect all third parties
  examine_third_parties(db, false);

  // now find and rank clusters
  examine_clusters();

  // graph largest cluster
  DashbotCluster *cluster = rbTreeLastValue(clusters);
  graph_cluster(cluster, "largest.dot");

  RbTree *node_list = sort_third_parties(db);

  rbTreeFree(node_list);
  rbTreeFree(clusters);
  hashtable_destroy(sites, 2);
  int count = free_all_nodes();
  printf("Found %d sites including 3rd parties\n", count);
  return 0;
}

