
/*
 *      gml.c -- generate gml file from dashboard data
 *
 *      Copyright (c) 2011 Dave Raggett
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>   // for strlen
#include <strings.h>
#include <sqlite3.h>

#include "hashtable.h"
#include "analyst.h"
#include "top_sites.h"

#define GML_FILE "dashboard.gml"

struct hashtable *gml_nodes = NULL;

static long int get_node(FILE *fp, char *host)
{
  static long int n = 0;
  long int id = (long int)hashtable_search(gml_nodes, host);

  if (id == 0)
  {
    id = ++n;
    host = strdup(host);
    assert(host);
    assert(hashtable_insert(gml_nodes, host, (void *)id, 0));

    fprintf(fp, "  node [ id %ld label \"%s\" ]\n",  id, host);
  }

  return id;
}

// 1st arg is context passed in 4th arg of sqlite3_exec
// 2nd arg is number of columns
// 3rd arg is an array of pointers to column values
// 4th arg is an array of pointers to column names
// note: column value may be null so check for null pointer
static int arcs(void *fp, int argc, char **argv, char **azColName)
{
  char *page_host = argv[0];
  char *third_party = argv[1];
  char *offsite = argv[2];

  int flag;
  sscanf(offsite, "%d", &flag);

  long int host_id = get_node((FILE *)fp, page_host);
  long int third_pty_id = get_node((FILE *)fp, third_party);

  fprintf(fp, "    edge [ source %ld target %ld ]\n", host_id, third_pty_id);

  return 0;
}

static void third_parties(FILE *fp, sqlite3 *db, char *host)
{
  char sql[1024];
  char *zErrMsg = 0;

  sprintf(sql,
      "SELECT page_host,third_party,offsite FROM parties WHERE page_host = \"www.%s\"",
      host);

  int rc = sqlite3_exec(db, sql, arcs, fp, &zErrMsg);

  if( rc != SQLITE_OK )
  {
    fprintf(stderr, "SQL error: %s\n", zErrMsg);
    sqlite3_free(zErrMsg);
  }
}

static void graph_gml_nodes(FILE *fp, sqlite3 *db)
{
  int i;
  SiteDesc *desc;

  for (i = 0; i < NUM_SITES; ++i)
  {
    desc = top_sites + i;
    get_node(fp, desc->domain);
    third_parties(fp, db, desc->domain);
  }
}

static int compare_strings(void *key1, void *key2)
{
  return !strcmp((char *)key1, (char *)key2);
}

void gml(sqlite3 *db)
{
  gml_nodes = create_hashtable(8192, oat_hash2, compare_strings);

  if (!gml_nodes)
  {
    fprintf(stderr, "Can't create hash table\n");
    sqlite3_close(db);
    exit(1);
  }

  char *name = GML_FILE;
  FILE *fp = fopen(name, "w+");

  if (!fp)
  {
    fprintf(stderr, "Can't open output file %s\n", name);
    sqlite3_close(db);
    exit(1);
  }

  fprintf(fp, "graph [\n");

  graph_gml_nodes(fp, db);
  fprintf(fp, "]\n");

  fclose(fp);
  fprintf(stderr, "written gml graph to %s\n", name);
  hashtable_destroy(gml_nodes, 0);
}

