/*
 *      rbtree.h -- Red Black balanced tree library
 *
 *      Copyright (c) 2008 Dave Raggett
 *      based upon code by Julienne Walker
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifndef RBTREE_H
#define RBTREE_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stddef.h>
#include <stdlib.h>

typedef void *RbKey;   // e.g. memory address of object
typedef void *RbValue;
typedef void *RbContext;
typedef unsigned long RbSize;  // long for compatibility with LP64

typedef struct _rbnode
{
    int red;     // Color (1=red, 0=black)
    RbKey key; // same size as void *
    RbValue value;  // User-defined content
    struct _rbnode *link[2]; // Left (0) and right (1) links
} RbNode;

// create new value in red-black tree node
typedef void (*RbNewFn) ( RbNode *node, RbKey key, RbValue value );

// for updating existing data in red-black tree node
typedef void  (*RbUpdateFn) ( RbNode *node, RbValue value );

// for freeing data stored in red-black tree node
typedef void  (*RbEraseFn) ( RbKey key, RbValue value );

typedef int (*RbCmpKeyFn) (RbKey key1, RbKey key2);

// user supplied function for apply mechanism
typedef void (*RbApplyFn)(RbKey key, RbValue value, RbContext context);

typedef struct _rbtree
{
    RbNewFn newNode;
    RbUpdateFn updateNode;
    RbEraseFn eraseNode;
    RbCmpKeyFn compareKeys;
    RbNode *root;
} RbTree;

// red black tree functions
RbTree *rbTreeNew(RbNewFn newNode, RbUpdateFn updateNode,
                   RbEraseFn eraseNode,  RbCmpKeyFn compareKeys);

void rbTreeNewEmbedded(RbTree *tree, RbNewFn newNode, RbUpdateFn updateNode,
                        RbEraseFn eraseNode,  RbCmpKeyFn compareKeys);

void rbTreeFree(RbTree *tree);

void rbTreeFreeEmbedded(RbTree *tree);


RbValue rbTreeFindKey ( RbTree *tree, RbKey key );
void rbTreeInsertKey ( RbTree *tree, RbKey key, RbValue data );
void rbTreeRemoveKey ( RbTree *tree, RbKey key );

RbSize rbTreeSize ( RbTree *tree ); // number of nodes in tree
int rbTreeHasOneEntry ( RbTree *tree ); // true of tree has just one node

RbValue rbTreeRootValue(RbTree *tree);
RbValue rbTreeFirstValue(RbTree *tree);
RbValue rbTreeLastValue(RbTree *tree);

// calls user defined func on all nodes with given context
void rbTreeApplyLR(RbTree *tree, RbApplyFn func, RbContext context);
void rbTreeApplyRL(RbTree *tree, RbApplyFn func, RbContext context);

 #ifdef __cplusplus
 }
 #endif

#endif // RBTREE_H
